import numpy as np
from skimage import draw, transform, io
from shapely import geometry
from matplotlib import pyplot as plt
import os
import json
import pathlib
import pandas as pd
import pingouin as pg
import scipy.stats
import math

# paths
DIR_SRC = pathlib.Path(__file__).parent.absolute()
FILE_ASSESSMENTS = os.path.join(DIR_SRC, 'assessments.json')
FILE_USERS = os.path.join(DIR_SRC, 'users.json')
FILE_IMAGES = os.path.join(DIR_SRC, 'images.json')
DIR_IMAGES = os.path.join(DIR_SRC, '..', 'images')
DIR_DATAFRAMES = os.path.join(DIR_SRC, '..', 'dataframes')

# dataset constants
IMG_SHAPE = np.array([720, 720])
SCORE_MAX = 5
RATINGS = {
    '80-100% readable': 5,
    '60-80% readable': 4,
    '40-60% readable': 3,
    '20-40% readable': 2,
    '0-20% readable': 1
}
LANGUAGES = [
    "Armenian",
    "Georgian",
    "German",
    "Gothic",
    "Greek",
    "Latin",
    "Ottoman",
    "Slavonic"
]

# constants that could be changed custom analyses
DEFAULT_SUPERPIXEL_SIZE = 24


class Maptypes:
    '''
    Basically enumerates the types of maps that can be produced for a given image by the function make_maps
    '''
    MEAN = 'mean_score_maps'            # mean scores
    STD = 'std_maps'                    # standard deviations of scores
    FOREGROUND = 'foreground_maps'      # True for pixels/units that are marked with an annotation by more than 50% of raters, False otherwise
    BORDERS = 'border_maps'             # see make_border_area_map
    all = [MEAN, STD, FOREGROUND, BORDERS]


class Data:
    '''
    Convenient access to the source data
    '''
    def __init__(self, omit_duplicates=True):
        '''
        :param omit_duplicates: if True, all duplicates used for intra-observer assessment are omitted.
                                Only set to False when using the Data for intra-observer assessment.
        '''
        with open(FILE_ASSESSMENTS) as f:
            self.assessments = json.load(f)
            if omit_duplicates:
                self.assessments = [a for a in self.assessments if a['instance'] == 0]
        with open(FILE_USERS) as f:
            self.users = json.load(f)
        with open(FILE_IMAGES) as f:
            self.images = json.load(f)

    def get_observations_dataframe(self, superpixel_size=1, omit_background_units=True, subsample=1.0):
        '''
        Creates a pandas dataframe of observations, or loads it if it already was created before.
        :param subsample: 1.0 (default): consider all units. 0 < subsamples < 1.0: return a random sample of
                            units of the given relative size. Useful for faster testing.
        :param omit_background_units: want to return only units rated as foreground by more than 50% of users?
        :return: pandas dataframe, where each row corresponds to a given unit observed by a given user.
        '''
        file = 'observations_sp%d.pkl' % superpixel_size
        fullfile = os.path.join(DIR_DATAFRAMES, file)
        if os.path.exists(fullfile):
            df = pd.read_pickle(fullfile)

        else:
            if not os.path.exists(DIR_DATAFRAMES):
                os.makedirs(DIR_DATAFRAMES)

            # enhance images with foreground maps and make them accessible by id in a dict
            images = {}
            for img in self.images:
                images[img['id']] = img
                print('making foreground map for image '+img['id'])
                images[img['id']]['fgmap'] = make_maps(
                    [a for a in self.assessments if a['img_id'] == img['id']],
                    types=[Maptypes.FOREGROUND],
                    superpixel_size=DEFAULT_SUPERPIXEL_SIZE
                )[Maptypes.FOREGROUND]

            rows = []
            for user in self.users:
                print('inserting observations of user ' + user['id'])
                for ass in [a for a in self.assessments if a['user_id'] == user['id']]:
                    scores = make_score_map(ass['annotations'], superpixel_size=superpixel_size)
                    scores = scores.ravel()
                    fg = images[ass['img_id']]['fgmap'].ravel()
                    cur_lang = images[ass['img_id']]['lang']
                    lang_fit = user['lang_' + cur_lang]
                    for i in range(scores.size):
                        rows.append({
                            # target
                            'score': scores[i],

                            # user
                            'user_id': user['id'],
                            'academic_level': user['academic_level'],
                            'msi_experience': user['msi_experience'],
                            'environment': user['environment'],

                            # observational unit
                            'region_id': ass['img_id'][0:2],
                            'img_id': ass['img_id'],
                            'unit_id': ass['img_id'] + '_' + str(i).zfill(3),
                            'lang': cur_lang,
                            'batch': images[ass['img_id']]['batch'],
                            'foreground': fg[i],

                            # assessment
                            'lang_fit': lang_fit,
                            'duration': ass['duration'],
                        })

            df = pd.DataFrame(rows)
            df.to_pickle(fullfile)

        if 0 < subsample < 1:
            regions = df['unit_id'].unique()
            regionsample = np.random.choice(regions, int(len(regions) * subsample))
            df = df.loc[df['unit_id'].isin(regionsample)]

        if omit_background_units:
            df = df.loc[df['foreground']]

        return df


def ann_clip(annotation):
    '''
    Clips annotation coordinates according to IMG_SHAPE.
    (Because the annotations could be dragged outside of the image in the assessment GUI)
    :param annotation: un-clipped annotation
    :return: clipped annotation
    '''
    xmin = np.clip(annotation['x'], 0, IMG_SHAPE[1]-1)
    ymin = np.clip(annotation['y'], 0, IMG_SHAPE[0]-1)
    xmax = np.clip(annotation['x'] + annotation['width'], 0, IMG_SHAPE[1])
    ymax = np.clip(annotation['y'] + annotation['height'], 0, IMG_SHAPE[0])
    annotation['x'] = xmin
    annotation['y'] = ymin
    annotation['width'] = xmax - xmin
    annotation['height'] = ymax - ymin
    return annotation


def ann_to_polygon(annotation):
    '''
    Converts annotation to shapely polygon
    :param annotation: annotation (as in the original data)
    :return: shapely polygon
    '''
    return geometry.box(
        annotation['x'],
        annotation['y'],
        annotation['x'] + annotation['width'] - 1,
        annotation['y'] + annotation['height'] - 1
    )


def whole_image_marked(annotation):
    '''
    Little helper to check if an annotation covers the whole image
    (created via 'mark whole image' in the assessment GUI)
    :param annotation: the annotation in question
    :return: True if this annotation marks the whole image, False otherwise.
    '''
    return annotation['x'] == annotation['y'] == 0 and all((annotation['height'], annotation['width']) == IMG_SHAPE)


def make_score_map(annotations, superpixel_size=1):
    '''
    Creates a map of pixel scores from a set of annotations from a single assessment (=one image rated by one observer).
    :param annotations: the input annotations
    :param superpixel_size: If set to a value > 1, pixels are grouped to square-shaped superpixels
                            with side-length superpixel_size. sizes of output maps are reduced accordingly.
    :return: Map of unit scores for a single observation. For overlapping observations, the maximum is taken.
    '''
    score_map = np.zeros(IMG_SHAPE)    #accumulates the scores from the anotations
    for ann in [ann_clip(a) for a in annotations]:
        rr, cc = draw.rectangle(start=(ann['y'], ann['x']),
                                extent=(ann['height'], ann['width']),
                                shape=IMG_SHAPE)
        score_map[rr, cc] = np.maximum(RATINGS[ann['rating']], score_map[rr, cc])
    if superpixel_size > 1:
        score_map = transform.downscale_local_mean(score_map, (superpixel_size, superpixel_size))
        score_map = np.round(score_map)
    return score_map


def make_border_area_map(assessments, t_iou=0.9, superpixel_size=1):
    '''
    We define border areas as the union of all symmetric differences of overlapping annotations with an intersection
    over union >= t_iou
    :param assessments: assessments of a single image
    :param t_iou: intersection-over-union threshold for a pair of annotations to be considered similar
    :param superpixel_size: If set to a value > 1, pixels are grouped to square-shaped superpixels
                            with side-length superpixel_size. sizes of output maps are reduced accordingly.
    :return: A map where border areas are 1 and other units are 0. With a superpixel_size>1, can contain values between
            0 and 1, due to local averaging.
    '''
    symdiffmap = np.zeros(IMG_SHAPE, dtype=bool)
    # check each assessment against each:
    for i in range(0, len(assessments)):
        for j in range(i + 1, len(assessments)):
            # for the current pair of assessments, compare all pairs of annotations:
            for ann_a in assessments[i]['annotations']:
                rect_a = ann_to_polygon(ann_clip(ann_a))
                for ann_b in assessments[j]['annotations']:
                    rect_b = ann_to_polygon(ann_clip(ann_b))
                    if rect_a.intersects(rect_b) and not(whole_image_marked(ann_a) and whole_image_marked(ann_b)):
                        # check if overlapping annotations are similar wrt. intersection over union
                        if rect_a.intersection(rect_b).area / rect_a.union(rect_b).area > t_iou:
                            # make symmetric difference by hand, because I don't know how to rasterize shapely polygons..
                            tmp_map = np.zeros(IMG_SHAPE)
                            for ann in [ann_a, ann_b]:
                                rr, cc = draw.rectangle(start=(ann['y'], ann['x']),
                                                        extent=(ann['height'], ann['width']),
                                                        shape=IMG_SHAPE)
                                tmp_map[rr, cc] = tmp_map[rr, cc]+1
                            symdiffmap |= (tmp_map == 1)
    symdiffmap = transform.downscale_local_mean(symdiffmap, (superpixel_size, superpixel_size))
    return symdiffmap


def make_maps(assessments: list, types=Maptypes.all, superpixel_size=1):
    '''
    Creates various maps for a single image from a list of assessments
    :param assessments: list of assessments FOR THE SAME IMAGE
    :param types: list of maptypes (members of Maptype), that indicates which maps should be made.
    :param superpixel_size: If set to a value > 1, pixels are grouped to square-shaped superpixels
                            with side-length superpixel_size. sizes of output maps are reduced accordingly.
    :return: a dict with types as keys and the respective maps as values
    '''

    score_stack = None
    for ass in assessments:
        cur_scoremap = make_score_map(ass['annotations'], superpixel_size)
        score_stack = cur_scoremap if score_stack is None else np.dstack((score_stack, cur_scoremap))

    # no assessment was given
    if score_stack is None:
        return None

    single_ass = len(score_stack.shape) < 3     # only single assessment given?
    maps = {}
    for mt in types:
        if mt == Maptypes.MEAN:
            maps[mt] = score_stack if single_ass else score_stack.mean(2)
        elif mt == Maptypes.STD:
            maps[mt] = np.zeros(score_stack.shape) if single_ass else score_stack.std(2)
        elif mt == Maptypes.FOREGROUND:
            maps[mt] = (score_stack > 0) if single_ass else ((score_stack > 0).sum(2) > (score_stack.shape[2] / 2))
        elif mt == Maptypes.BORDERS:
            maps[mt] = make_border_area_map(assessments, superpixel_size=superpixel_size)

    return maps


def write_all_maps_to_images(data: Data, types, superpixel_size=1):
    '''
    Make specified maps for all input images and write them to images.
    Normalization:
        mean scores:                [0,5]   -> [0, 255]
        stds:                       [0,2.5] -> [0,255]
        foreground & border maps:   [0,1]   -> [0,255]
    :param data: Data from which the images are generated
    :param types: list of maptypes (members of Maptype), that indicates which maps should be made.
    :param superpixel_size: If set to a value > 1, pixels are grouped to square-shaped superpixels
                            with side-length superpixel_size. sizes of output maps are reduced accordingly.
    '''

    # create output dirs
    for mt in types:
        d = os.path.join(DIR_IMAGES, mt)
        if not os.path.exists(d):
            os.makedirs(d)

    for img_id in [i['id'] for i in data.images]:
        print('making maps for image ' + img_id)
        cur_ass = [a for a in data.assessments if a['img_id'] == img_id]
        maps = make_maps(cur_ass, types=types, superpixel_size=superpixel_size)
        for mt in types:
            img = maps[mt]
            if mt == Maptypes.MEAN:
                img = img / SCORE_MAX
            if mt == Maptypes.STD:
                img = img / (SCORE_MAX / 2)
            io.imsave(os.path.join(DIR_IMAGES, mt, img_id + '.png'), np.uint8(img * 255))


def compute_iccs(df, filters={}):
    '''
    Computes intra-class correlation coefficients for the given DataFrame, using column 'unit_id' as the rated target,
    'score' as the rating value and 'user_id' as the rater.
    :param df: a pandas.DataFrame containing the observations.
    :param filters: a dict to set filters on columns, where keys are column names and values are the accepted values
                    for that column (can be a single value or a list of accepted values)
    :return: the icc results as returned by pingouin.intraclass_corr
    '''
    idx = np.ones(len(df.index), dtype=bool)
    descr = ''
    for k, v in filters.items():
        if isinstance(v, list):
            idx2 = np.zeros(len(df.index), dtype=bool)
            for val in v:
                idx2 |= (df[k] == val)
            idx &= idx2
            descr += (k + ' in ' + str(v) + ' ')
        else:
            idx &= (df[k] == v)
            descr += (k+'='+str(v)+' ')
    if descr == '':
        descr = 'all'

    try:
        icc = pg.intraclass_corr(df.loc[idx],
                                 targets='unit_id',
                                 raters='user_id',
                                 ratings='score',
                                 nan_policy='omit'
                                 )
        print('ICC for %s:' % descr)
        print(icc)
        return icc
    except:
        print('Not enough matching rows for ' + descr)
        return None
