# == USAGE EXAMPLES ==
# The following usage examples reproduce the published dataset and results described in the
# related paper. Be sure to set the the constant FRACTION_OF_DATA_TO_USE is set to 1.0 for 100% reproduction.
# All the functions are called at the bottom of the file.

from salami_proc import *
from matplotlib import pyplot as plt
from matplotlib.lines import Line2D
import seaborn as sns
sns.set(font_scale=1.5)
sns.set_style("white")


# FRACTION_OF_DATA_TO_USE is only relevant for global_histograms, inter_rater_agreement and plot_mean_effects
# SET IT TO 1.0 to reproduce results from paper.
# Leave it at a smaller value for faster results but larger uncertainties.
FRACTION_OF_DATA_TO_USE = 0.01

# create mean score maps and std ('uncertainty') maps as images, in full resolution
def make_mean_score_and_std_maps():
    data = Data(omit_duplicates=True)
    maptypes = [Maptypes.MEAN, Maptypes.STD]
    write_all_maps_to_images(data, maptypes, superpixel_size=1)

# get quartiles of time required to rate an image
def duration_quartiles():
    data = Data(omit_duplicates=True)
    durations = [a['duration'] for a in data.assessments]
    quartiles = np.quantile(durations, [0.25, 0.5, 0.75])
    print('image rating duration quartiles (ms): %d, %d, %d' % tuple(quartiles))

# screen observers, following ITU-R Recommendation BT.500. (2012).
def participant_screening():
    data = Data(omit_duplicates=True)

    # for each image, generate a map of mean scores and a map of 'margins', i.e. the allowed deviation from the mean
    # the margin is 2*std if the scores are normally distributed (Fisher Kurtosis between -1 and 1), and sqrt(2)*std otherwise.
    for img in data.images:
        print('Making stats maps for image '+ img['id'])
        score_stack = None
        for ass in [a for a in data.assessments if a['img_id'] == img['id']]:
            score_map = make_score_map(ass['annotations'], superpixel_size=DEFAULT_SUPERPIXEL_SIZE)
            score_stack = score_map if score_stack is None else np.dstack((score_stack, score_map))
        img['mean'] = score_stack.mean(2)
        std = score_stack.std(2)
        kurt = scipy.stats.kurtosis(score_stack, 2)
        is_normal = (-1 <= kurt) <= 1
        img['margin'] = np.zeros(is_normal.shape)
        img['margin'][is_normal] = std[is_normal] * 2
        img['margin'][np.logical_not(is_normal)] = std[np.logical_not(is_normal)] * math.sqrt(20)

    # now count the 'faults' of each participant and decide upon rejection
    for user_id in [u['id'] for u in data.users]:
        P = 0
        Q = 0
        totes = 0
        for img in data.images:
            asses = [a for a in data.assessments if a['user_id'] == user_id and a['img_id'] == img['id']]
            if asses == []:
                continue
            this_score = make_score_map(asses[0]['annotations'], superpixel_size=DEFAULT_SUPERPIXEL_SIZE)
            too_much = this_score > img['mean'] + img['margin']
            too_little = this_score < img['mean'] - img['margin']
            P += too_much.sum()
            Q += too_little.sum()
            totes += this_score.size

        rejected = (P + Q) / totes > 0.05 and abs((P - Q) / (P + Q)) < 0.3
        print('user %s: %s' % (user_id, 'REJECTED!' if rejected else 'OK'))

# make/load pandas dataframe of regular observations, including background. then use it to:
# - make a score histogram of all observations
# - make a 2d histogram of scores vs. stds per unit
def global_histograms():
    data = Data(omit_duplicates=True)
    df = data.get_observations_dataframe(superpixel_size=DEFAULT_SUPERPIXEL_SIZE,
                                         omit_background_units=False,
                                         subsample=FRACTION_OF_DATA_TO_USE)

    # somehow pyplot.hist doesn't do what i want here, so..
    bins = range(6)
    counts = []
    for b in bins:
        counts.append(df[df['score'] == b].shape[0])
    plt.bar(bins, np.array(counts) / df.shape[0])
    plt.xlabel('score')
    plt.ylabel('relative frequency')
    plt.xticks(bins)
    plt.show()

    means = df.groupby(['unit_id'])['score'].mean()
    stds = df.groupby(['unit_id'])['score'].std()
    plt.hist2d(means, stds, bins=5)
    plt.xlabel('mean score')
    plt.ylabel('standard deviation')
    plt.xticks(range(6))
    plt.colorbar()
    plt.show()

#compute intra-observer errors
def intra_rater_errors():
    data = Data(omit_duplicates=False)

    observer_errs = []
    for dup in [a for a in data.assessments if a['instance'] == 1]:
        orig = [a for a in data.assessments if a['user_id'] == dup['user_id'] and
                a['img_id'] == dup['img_id'] and
                a['instance'] == 0][0]
        score_map_dup = make_score_map(dup['annotations'], superpixel_size=DEFAULT_SUPERPIXEL_SIZE)
        score_map_orig = make_score_map(orig['annotations'], superpixel_size=DEFAULT_SUPERPIXEL_SIZE)

        errors = np.absolute(score_map_dup - score_map_orig).ravel()
        # uncomment th next to lines if you want to include all units:
        nonzeros = np.logical_or(score_map_dup > 0, score_map_orig > 0).ravel()
        errors = errors[nonzeros]

        observer_errs.append(errors)

    observer_errs = np.concatenate(observer_errs)
    mean = observer_errs.mean()
    median = np.median(observer_errs)
    bins = range(6)
    counts = []
    for i in bins:
        subset = observer_errs[observer_errs == i]
        counts.append(len(subset))
    print('Mean=%f, Median=%f' % (mean, median))
    print(counts)

    plt.figure(figsize=(10, 4))
    rects = plt.barh(bins, np.array(counts) / len(observer_errs))
    # labels on top of bars:
    for rect in rects:
        width = rect.get_width()
        plt.annotate('%.4f' % width,
                     xy=(width, rect.get_y() + rect.get_height() / 2),
                     xytext=(3, 0),  # 3 points horizontal offset
                     textcoords="offset points",
                     ha='left', va='center')
    plt.xlim(10 ** (-3.2), 10 ** (0.4))
    plt.xscale('log')
    plt.yticks(bins)
    plt.xlabel('relative error counts')
    plt.ylabel('error magnitute')
    plt.show()

# make/load pandas dataframe of all regular observations, without background
# and compute inter-observer agreement with ICC (for all participants and for various groupings)
# this is quite time-consuming when using the whole dataset; set the subsample parameter of
# get_observations_dataframe to a lower value (e.g. 0.1) for a faster test (with almost identical results)
def inter_rater_agreement():
    data = Data(omit_duplicates=True)
    df = data.get_observations_dataframe(superpixel_size=DEFAULT_SUPERPIXEL_SIZE,
                                         omit_background_units=True,
                                         subsample=FRACTION_OF_DATA_TO_USE)

    #icc for all observations
    compute_iccs(df)

    #icc by environment
    compute_iccs(df, {'environment': 'lab'})
    compute_iccs(df, {'environment': 'online'})

    #icc by msi experience
    compute_iccs(df, {'msi_experience': [0, 1, 2]})
    compute_iccs(df, {'msi_experience': 3})

    #icc by language expertises (with nice plot)
    n_levels = 4
    plt.figure(figsize=[10,5])
    for lang, i in zip(['Latin', 'Greek', 'Slavonic'], range(3)):
        iccs = []
        for lang_fit in range(n_levels):
            icc_all = compute_iccs(df, {'lang': lang, 'lang_fit': lang_fit})
            if icc_all is not None:
                icc = [icc_all['ICC'][1], icc_all['CI95%'][1]]
                iccs.append(icc)
        est = np.array([icc[0] for icc in iccs if not np.isnan(icc[0])])
        x = np.array([i for icc, i in zip(iccs, range(len(iccs))) if not np.isnan(icc[0])])
        errneg = est-np.array([icc[1][0] for icc in iccs if not np.isnan(icc[0])])
        errpos = np.array([icc[1][1] for icc in iccs if not np.isnan(icc[0])])-est
        p = plt.errorbar(
            y=est,
            x=x+((i-1)/20),
            yerr=[errneg, errpos],
            label=lang,
            marker='D', capsize=10, markersize=10
        )
    legend_elements = [Line2D([0], [0], color='w', marker='D', markerfacecolor=plt.rcParams['axes.prop_cycle'].by_key()['color'][0], markersize=10, label='Latin'),
                       Line2D([0], [0],  color='w', marker='D', markerfacecolor=plt.rcParams['axes.prop_cycle'].by_key()['color'][1], markersize=10, label='Greek'),
                       Line2D([0], [0],  color='w', marker='D', markerfacecolor=plt.rcParams['axes.prop_cycle'].by_key()['color'][2], markersize=10, label='Slavonic')
                       ]
    plt.legend(handles=legend_elements, loc='lower left')
    plt.xticks(range(n_levels))
    plt.xlabel('expertise level')
    plt.ylabel('ICC(2,1)')
    plt.show()

# plot mean scores grouped by various parameters. This is rediculously slow when using the full dataset.
# just set subsample to a smaller value in Data.get_observations_dataframe() to speed thing up
# (but then the reduced dataframe might have to be made first)
def plot_mean_effects():
    df = Data(omit_duplicates=True).get_observations_dataframe(superpixel_size=DEFAULT_SUPERPIXEL_SIZE,
                                                               omit_background_units=True,
                                                               subsample=FRACTION_OF_DATA_TO_USE)
    sns.lineplot('lang_fit', 'score', data=df)
    plt.ylabel('mean score')
    plt.xticks([0,1,2,3])
    plt.xlabel('language fit')
    plt.show()
    sns.lineplot('academic_level', 'score', data=df)
    plt.xticks([1,2,3,4])
    plt.xlabel('academic level')
    plt.ylabel('mean score')
    plt.show()
    sns.lineplot('msi_experience', 'score', data=df)
    plt.xlabel('SMI exposure')
    plt.ylabel('mean score')
    plt.xticks([1,2,3])
    plt.show()
    sns.lineplot('environment', 'score', data=df)
    plt.ylabel('mean score')
    plt.show()
    plt.figure(figsize=[10, 4])
    sns.barplot('user_id', 'score', data=df)
    plt.xlabel('participant id')
    plt.ylabel('mean score')
    plt.show()

# save a 50% subsample of observational units, without background.
# this was used in R to fit the linear mixed effects models. 50% sample because then you could theoretically
# test the model with the rest
def export_dataframe_for_r():
    np.random.seed([ord(c) for c in 'salami'])  # seed for the reproducibility
    df = Data(omit_duplicates=True).get_observations_dataframe(superpixel_size=DEFAULT_SUPERPIXEL_SIZE,
                                                               omit_background_units=True,
                                                               subsample=0.5)
    df.to_pickle(os.path.join(DIR_DATAFRAMES, 'observations_sp%d_0.5.pkl' % DEFAULT_SUPERPIXEL_SIZE))

# compare the differences in standard deviations in border areas and other areas
def compare_stds_border_areas_vs_rest():
    data = Data(omit_duplicates=True)
    stds_border = []
    stds_other = []
    for img_id in [i['id'] for i in data.images]:
        print('processing images '+img_id)
        maps = make_maps([a for a in data.assessments if a['img_id'] == img_id],
                         [Maptypes.STD, Maptypes.BORDERS],
                         superpixel_size=DEFAULT_SUPERPIXEL_SIZE)
        borderunits = np.round(maps[Maptypes.BORDERS]).astype(bool).ravel()
        stds = maps[Maptypes.STD].ravel()
        stds_border.append(stds[borderunits])
        stds_other.append(stds[np.logical_not(borderunits)])

    stds_border = np.concatenate(stds_border)
    stds_other = np.concatenate(stds_other)

    plt.figure(figsize=[10, 6])
    for name, values in zip(['border areas', 'other areas'], [stds_border, stds_other]):
        print(name+':')
        print('mean='+str(values.mean()))
        print('std=' + str(values.std()))
        k2, p = scipy.stats.normaltest(values)
        print('normaltest=' + str(p))
        nbins = 20
        sns.distplot(values, label=name, bins=nbins)

    print('T-Test: p='+str(scipy.stats.mannwhitneyu(stds_other, stds_border, alternative='less')))
    print('Ratio fraction of border areas: ' + str(stds_border.size/(stds_border.size+stds_other.size)))
    plt.legend()
    plt.show()


# HERE THE FUNCTIONS ARE CALLED.
# You might want to comment/remove some of those lines:
if __name__ == "__main__":
    make_mean_score_and_std_maps()
    # duration_quartiles()
    # participant_screening()
    # global_histograms()
    # intra_rater_errors()
    # inter_rater_agreement()
    # plot_mean_effects()
    # export_dataframe_for_r()
    # compare_stds_border_areas_vs_rest()
